from metasequoia_connector.connector.mysql import MysqlConnector
from metasequoia_connector.node import DSMetaInstance

__all__ = ["DSMetaConnector"]


class DSMetaConnector(MysqlConnector):
    """海豚调度元数据连接器"""

    def __init__(self, dolphin_scheduler_meta_info: DSMetaInstance):
        super().__init__(dolphin_scheduler_meta_info,
                         schema=dolphin_scheduler_meta_info.db)
        self._dolphin_scheduler_meta_info = dolphin_scheduler_meta_info



"""
otssql 工具函数
"""

from typing import Any, Dict, Tuple

import otssql
from metasequoia_connector.manager.connect_manager import ConnectManager


def conn_select_all_as_dict(conn: otssql.Connection, sql: str) -> Tuple[Dict[str, Any], ...]:
    """使用 sql 查询 conn 连接的 MySQL，并将结果作为 dict 的列表返回"""
    with conn.cursor(otssql.cursor.DictCursor) as cursor:
        cursor.execute(sql)
        return cursor.fetchall()


def select_all_as_dict(manager: ConnectManager, ots_name: str, sql: str, **params):
    with manager.otssql_connect(ots_name, **params) as ots_conn:
        return conn_select_all_as_dict(conn=ots_conn, sql=sql)


def conn_select_one_as_dict(conn: otssql.Connection, sql: str) -> Dict[str, Any]:
    with conn.cursor(otssql.cursor.DictCursor) as cursor:
        cursor.execute(sql)
        return cursor.fetchone()


def select_one_as_dict(manager: ConnectManager, ots_name: str, sql: str):
    with manager.otssql_connect(ots_name) as ots_conn:
        return conn_select_one_as_dict(conn=ots_conn, sql=sql)


def conn_execute_and_commit(conn: otssql.Connection, sql: str) -> int:
    """执行 SQL 语句，并 commit"""
    with conn.cursor(otssql.cursor.DictCursor) as cursor:
        result = cursor.execute(sql)
    conn.commit()
    return result


def execute_and_commit(manager: ConnectManager, ots_name: str, sql: str) -> int:
    with manager.otssql_connect(ots_name) as ots_conn:
        return conn_execute_and_commit(conn=ots_conn, sql=sql)



"""
Hive 连接器
"""

import random
from typing import Optional

from pyhive import hive

from metasequoia_connector.connector.ssh_tunnel import create_ssh_tunnel
from metasequoia_connector.node import HiveInstance

__all__ = ["HiveConnector"]


class HiveConnector:
    def __init__(self,
                 hive_instance: HiveInstance,
                 schema: Optional[str] = None) -> None:
        """MySQL 连接的构造方法

        Parameters
        ----------
        hive_instance : RdsInstance
            MySQL 实例的配置
        schema : Optional[str], default = None
            数据库名称
        """
        self.hive_instance_info = hive_instance  # MySQl 实例的配置
        self.schema = schema  # 数据库

        # 初始化 Hive 连接和 SSH 隧道连接
        self.hive_conn = None
        self.ssh_tunnel_conn = None

    def __enter__(self):
        """在进入 with as 语句的时候被 with 调用，返回值作为 as 后面的变量

        因为 pyhive 不支持连接多个 Hive Client 的集群，因此在其中随机选择
        """

        choose_host = random.choice(self.hive_instance_info.hosts)

        ssh_tunnel_info = self.hive_instance_info.ssh_tunnel

        if ssh_tunnel_info is not None:
            # 启动 SSH 隧道
            self.ssh_tunnel_conn = create_ssh_tunnel(
                ssh_tunnel_info,
                remote_bind_address=(choose_host, self.hive_instance_info.port)
            )

            # 更新 MySQL 连接信息，令 MySQL 连接到 SSH 隧道
            host = "127.0.0.1"
            port = self.ssh_tunnel_conn.local_bind_port
        else:
            host = choose_host
            port = self.hive_instance_info.port

        self.hive_conn = hive.Connection(host=host, port=port, username=self.hive_instance_info.username)

        return self.hive_conn

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.hive_conn is not None:
            self.hive_conn.close()
        if self.ssh_tunnel_conn is not None:
            self.ssh_tunnel_conn.close()



"""
Hive 连接器节点
"""

import dataclasses
from typing import List, Optional

from metasequoia_connector.node.ssh_tunnel import SshTunnel

__all__ = ["HiveInstance", "HiveTable"]


@dataclasses.dataclass(slots=True)
class HiveInstance:
    """Hive 实例"""

    hosts: List[str] = dataclasses.field(kw_only=True)
    port: int = dataclasses.field(kw_only=True)
    username: Optional[str] = dataclasses.field(kw_only=True, default=None)
    ssh_tunnel: Optional[SshTunnel] = dataclasses.field(kw_only=True, default=None)

    def set_hive_username(self, username: str) -> None:
        self.username = username


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class HiveTable:
    """Hive 表"""

    instance: HiveInstance = dataclasses.field(kw_only=True)
    schema: str = dataclasses.field(kw_only=True)
    table: str = dataclasses.field(kw_only=True)



"""
Kafka 连接器：基于 kafka-python
"""

from typing import List, Optional

from kafka import KafkaConsumer, KafkaProducer
from kafka.admin import KafkaAdminClient

from metasequoia_connector.connector.ssh_tunnel import create_ssh_tunnel
from metasequoia_connector.node import KafkaServer, KafkaTopicsGroup

__all__ = ["ConnKafkaAdminClient", "ConnKafkaConsumer", "ConnKafkaProducer"]


class ConnKafkaAdminClient:
    """根据 KafkaServer 对象创建 kafka-python 的 KafkaAdminClient 对象"""

    def __init__(self, kafka_server: KafkaServer):
        self.kafka_server = kafka_server  # MySQl 实例的配置

        # 初始化 MySQL 连接和 SSH 隧道连接
        self.kafka_admin_client = None
        self.ssh_tunnel_conn = None

    def __enter__(self):
        """在进入 with as 语句的时候被 with 调用，返回值作为 as 后面的变量"""
        if self.kafka_server.ssh_tunnel is not None:
            # 计算远端的地址立列表
            remote_bind_addresses = [(host_port.host, host_port.port)
                                     for host_port in self.kafka_server.get_host_list()]

            # 启动 SSH 隧道
            self.ssh_tunnel_conn = create_ssh_tunnel(
                self.kafka_server.ssh_tunnel,
                remote_bind_addresses=remote_bind_addresses
            )

            # 更新 Kafka 集群连接信息，令 Kafka 集群连接到 SSH 隧道
            if len(remote_bind_addresses) > 1:
                addresses = [f"127.0.0.1:{local_bind_port}"
                             for local_bind_port in self.ssh_tunnel_conn.local_bind_ports]
            else:
                addresses = [f"127.0.0.1:{self.ssh_tunnel_conn.local_bind_port}"]
        else:
            addresses = [f"{host_port.host}:{host_port.port}" for host_port in self.kafka_server.get_host_list()]

        # 启动 Kafka 集群连接
        self.kafka_admin_client = KafkaAdminClient(bootstrap_servers=addresses)

        return self.kafka_admin_client

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.kafka_admin_client is not None:
            self.kafka_admin_client.close()
        if self.ssh_tunnel_conn is not None:
            self.ssh_tunnel_conn.close()


class ConnKafkaConsumer:
    """根据 KafkaServer 对象创建 kafka-python 的 KafkaConsumer 对象"""

    def __init__(self, kafka_server: KafkaServer,
                 topics: Optional[List[str]] = None,
                 group_id: Optional[str] = None):
        self.kafka_server = kafka_server  # MySQl 实例的配置
        self.topics = topics
        self.group_id = group_id

        # 初始化 MySQL 连接和 SSH 隧道连接
        self.kafka_consumer = None
        self.ssh_tunnel_conn = None

    @staticmethod
    def by_kafka_topic_group(kafka_topics_group: KafkaTopicsGroup):
        """根据 KafkaTopicGroup 对象构造 KafkaConsumer 对象"""
        return ConnKafkaConsumer(
            kafka_server=kafka_topics_group.kafka_server,
            topics=kafka_topics_group.topics,
            group_id=kafka_topics_group.group_id
        )

    def __enter__(self):
        """在进入 with as 语句的时候被 with 调用，返回值作为 as 后面的变量"""
        if self.kafka_server.ssh_tunnel is not None:
            # 计算远端的地址立列表
            remote_bind_addresses = [(host_port.host, host_port.port)
                                     for host_port in self.kafka_server.get_host_list()]

            # 启动 SSH 隧道
            self.ssh_tunnel_conn = create_ssh_tunnel(
                self.kafka_server.ssh_tunnel,
                remote_bind_addresses=remote_bind_addresses
            )

            # 更新 Kafka 集群连接信息，令 Kafka 集群连接到 SSH 隧道
            if len(remote_bind_addresses) > 1:
                addresses = [f"127.0.0.1:{local_bind_port}"
                             for local_bind_port in self.ssh_tunnel_conn.local_bind_ports]
            else:
                addresses = [f"127.0.0.1:{self.ssh_tunnel_conn.local_bind_port}"]
        else:
            addresses = [f"{host_port.host}:{host_port.port}" for host_port in self.kafka_server.get_host_list()]

        # 启动 Kafka 集群连接
        if self.topics is not None:
            self.kafka_consumer = KafkaConsumer(*self.topics,
                                                bootstrap_servers=addresses,
                                                group_id=self.group_id,
                                                auto_offset_reset="latest",
                                                api_version=(0, 11))
        else:
            self.kafka_consumer = KafkaConsumer(bootstrap_servers=addresses,
                                                group_id=self.group_id,
                                                auto_offset_reset="latest",
                                                api_version=(0, 11))

        return self.kafka_consumer

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.kafka_consumer is not None:
            self.kafka_consumer.close()
        if self.ssh_tunnel_conn is not None:
            self.ssh_tunnel_conn.close()


class ConnKafkaProducer:
    """根据 KafkaServer 对象创建 kafka-python 的 KafkaProducer 对象"""

    def __init__(self, kafka_server: KafkaServer):
        self.kafka_server = kafka_server  # MySQl 实例的配置

        # 初始化 MySQL 连接和 SSH 隧道连接
        self.kafka_producer = None
        self.ssh_tunnel_conn = None

    def __enter__(self):
        """在进入 with as 语句的时候被 with 调用，返回值作为 as 后面的变量"""
        if self.kafka_server.ssh_tunnel is not None:
            # 计算远端的地址立列表
            remote_bind_addresses = [(host_port.host, host_port.port)
                                     for host_port in self.kafka_server.get_host_list()]

            # 启动 SSH 隧道
            self.ssh_tunnel_conn = create_ssh_tunnel(
                self.kafka_server.ssh_tunnel,
                remote_bind_addresses=remote_bind_addresses
            )

            # 更新 Kafka 集群连接信息，令 Kafka 集群连接到 SSH 隧道
            if len(remote_bind_addresses) > 1:
                addresses = [f"127.0.0.1:{local_bind_port}"
                             for local_bind_port in self.ssh_tunnel_conn.local_bind_ports]
            else:
                addresses = [f"127.0.0.1:{self.ssh_tunnel_conn.local_bind_port}"]
        else:
            addresses = [f"{host_port.host}:{host_port.port}" for host_port in self.kafka_server.get_host_list()]

        # 启动 Kafka 集群连接
        self.kafka_producer = KafkaProducer(bootstrap_servers=addresses,
                                            batch_size=1048576,
                                            linger_ms=50)

        return self.kafka_producer

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.kafka_producer is not None:
            self.kafka_producer.close()
        if self.ssh_tunnel_conn is not None:
            self.ssh_tunnel_conn.close()



"""
通用工具类
"""

from typing import Optional

__all__ = ["HostPort"]


class HostPort:
    """服务器地址和端口"""

    def __init__(self, host: str, port: Optional[int]):
        self._host = host
        self._port = port

    @staticmethod
    def create_by_url(url: str) -> "HostPort":
        if ":" not in url:
            return HostPort(url.strip(), None)
        else:
            host, port = url.split(":")
            return HostPort(host.strip(), int(port))

    @property
    def host(self) -> str:
        return self._host

    @property
    def port(self) -> Optional[int]:
        return self._port

    def __hash__(self):
        return hash((self._host, self._port))

    def __eq__(self, other):
        return (isinstance(other, HostPort) and
                self._host == other._host and
                self._port == other._port)



import json
import os
import typing
from typing import Any, Dict, List, Optional, Tuple

import dbutils.pooled_db
import redis

from metasequoia_connector.connector import DSMetaConnector
from metasequoia_connector.connector.common import WrapConnector
from metasequoia_connector.connector.kafka import ConnKafkaConsumer
from metasequoia_connector.connector.mysql import MysqlConnectionPool, MysqlConnector
from metasequoia_connector.connector.ots import OTSConnector
from metasequoia_connector.connector.redis import RedisConnectionPool, RedisConnector
from metasequoia_connector.node import (DSMetaInstance, HiveInstance, KafkaServer, KafkaTopic, KafkaTopicsGroup,
                                        MysqlInstance, OTSInstance, RedisDatabase, RedisInstance, SshTunnel)

__all__ = ["ConnectManager"]

# 配置文件的默认环境变量名称
DEFAULT_METASEQUOIA_CONNECTOR_CONFIG_NAME = "METASEQUOIA_CONNECTOR_CONFIG"


# TODO 将配置输出到文件：https://www.cnblogs.com/wengzx/p/18019494


class ConnectManager:
    """连接管理器"""

    ENCODING = "UTF-8"  # 编码格式

    def __init__(self, configuration: Dict[str, Any], mode: str = "dev"):
        self._configuration = configuration
        self._mode = mode

        # 缓存的连接池
        self._redis_pool_dict: Dict[str, RedisConnectionPool] = {}
        self._mysql_pool_dict: Dict[Tuple[str, Optional[str]], MysqlConnectionPool] = {}
        self._ots_conn_dict: Dict[str, OTSConnector] = {}

    @classmethod
    def from_config_file(cls, config_path: str, mode: str = "dev") -> "ConnectManager":
        """根据配置文件路径，并加载配置文件"""
        with open(config_path, "r", encoding=ConnectManager.ENCODING) as file:
            configuration = json.load(file)
        return ConnectManager(
            configuration=configuration,
            mode=mode
        )

    @classmethod
    def from_environment(cls, environ_name: Optional[str] = None, mode: str = "dev") -> "ConnectManager":
        """从环境变量中读取配置文件路径，并加载配置文件

        Parameters
        ----------
        environ_name : Optional[str], default = None
            环境变量名称，如果为空则使用默认名称
        mode : str, default = "dev"
            使用模式（会根据模式读取配置文件中的配置信息）
        """
        # 从环境变量中读取文件路径
        if environ_name is None:
            environ_name = DEFAULT_METASEQUOIA_CONNECTOR_CONFIG_NAME
        config_path = os.environ.get(environ_name)

        # 加载配置文件
        with open(config_path, "r", encoding=ConnectManager.ENCODING) as file:
            configuration = json.load(file)

        return ConnectManager(
            configuration=configuration,
            mode=mode
        )

    def __hash__(self):
        return self.get_hash(self._configuration)

    @staticmethod
    def get_hash(obj: object):
        if isinstance(obj, dict):
            return hash(tuple([(key, ConnectManager.get_hash(value)) for key, value in obj.items()]))
        if isinstance(obj, list):
            return hash(tuple([ConnectManager.get_hash(elem) for elem in obj]))
        return hash(obj)

    def __eq__(self, other: str) -> bool:
        if not isinstance(other, ConnectManager):
            return False
        return self._configuration == other._configuration

    # ------------------------------ 读取 MySQL 相关配置 ------------------------------

    def mysql_list(self) -> List[str]:
        """获取 MySQL 列表"""
        return self._get_name_list("MySQL")

    def mysql_dict(self, mysql_name: str) -> Dict[str, Any]:
        """获取 MySQL 信息"""
        return self._confirm_params(
            section="MySQL",
            config=self._get_section("MySQL", mysql_name),
            params=["host", "port", "user", "passwd"]
        )

    def mysql_obj(self, name: str) -> MysqlInstance:
        """获取 MysqlInstance 对象"""
        mysql_info = self.mysql_dict(name)
        ssh_tunnel = self.ssh_tunnel_obj(mysql_info["use_ssh"]) if mysql_info.get("use_ssh") else None
        return MysqlInstance(host=mysql_info["host"],
                             port=mysql_info["port"],
                             user=mysql_info["user"],
                             passwd=mysql_info["passwd"],
                             ssh_tunnel=ssh_tunnel)

    def mysql_refer_name(self, name: str) -> str:
        """获取 MySQL 的名称"""
        return self._configuration["MySQL"][name].get("_name", "")

    def mysql_refer_type(self, name: str) -> str:
        """获取 MySQL 的类型"""
        return self._configuration["MySQL"][name].get("_type", "")

    def mysql_refer_type_enum(self):
        """获取 MySQL 的 _type 属性的枚举值"""
        return {self.mysql_refer_type(mysql_name) for mysql_name in self.mysql_list()}

    def mysql_start_pool(self, mysql_name: str, schema_name: Optional[str] = None) -> MysqlConnectionPool:
        """根据 MySQL 的配置项名称，创建连接池

        Parameters
        ----------
        mysql_name : str
            MySQL 配置项的名称
        schema_name : Optional[str] = None
            数据库名称

        Returns
        -------
        MysqlConnectionPool
            MySQL 连接池对象
        """
        if (mysql_name, schema_name) in self._mysql_pool_dict:
            raise KeyError(f"重复启动 MySQL 连接池: ({mysql_name}, {schema_name})")
        mysql_instance = self.mysql_obj(mysql_name)
        self._mysql_pool_dict[(mysql_name, schema_name)] = MysqlConnectionPool(mysql_instance=mysql_instance,
                                                                               schema=schema_name)
        return self._mysql_pool_dict[(mysql_name, schema_name)]

    def mysql_close_pool(self, mysql_name: str, schema_name: Optional[str] = None) -> None:
        """根据 MySQL 的配置项名称，关闭连接池"""
        if (mysql_name, schema_name) not in self._mysql_pool_dict:
            raise KeyError(f"要关闭的 MySQL 连接池不存在: ({mysql_name}, {schema_name})")
        self._mysql_pool_dict[(mysql_name, schema_name)].close()  # 关闭连接池
        del self._mysql_pool_dict[(mysql_name, schema_name)]  # 从缓存连接池中移除连接池

    def mysql_connect(self, mysql_name: str, schema_name: Optional[str] = None, use_pool: bool = True):
        """根据 MySQL 配置项中名称，获取连接（当 use_pool 为 True 时，从连接池中获取；否则，创建新连接）"""
        if use_pool is True:
            if (mysql_name, schema_name) not in self._mysql_pool_dict:
                self.mysql_start_pool(mysql_name, schema_name)
            connection_pool = self._mysql_pool_dict[(mysql_name, schema_name)]
            connection_pool = typing.cast(dbutils.pooled_db.PooledDB, connection_pool)
            return WrapConnector(connection_pool.connection())
        return MysqlConnector(mysql_instance=self.mysql_obj(mysql_name), schema=schema_name)

    def mysql_new_connect(self, name: str, schema_name: Optional[str] = None):
        """根据实例名称，创建 MySQL 连接器"""
        return MysqlConnector(mysql_instance=self.mysql_obj(name), schema=schema_name)

    # ------------------------------ 读取 Ssh_Tunnel 相关配置 ------------------------------

    def ssh_tunnel_list(self):
        """获取 SSH 列表"""
        return self._get_name_list("Ssh_Tunnel")

    def ssh_tunnel_dict(self, ssh_tunnel_name: str) -> Dict[str, Any]:
        """获取 SSH 信息"""
        return self._confirm_params(
            section="Ssh_Tunnel",
            config=self._get_section("Ssh_Tunnel", ssh_tunnel_name),
            params=["host", "port"]
        )

    def ssh_tunnel_obj(self, ssh_tunnel_name: str) -> SshTunnel:
        """获取 SshTunnel 对象"""
        ssh_info = self.ssh_tunnel_dict(ssh_tunnel_name)
        return SshTunnel(host=ssh_info["host"],
                         port=ssh_info["port"],
                         username=ssh_info["username"],
                         password=ssh_info.get("password"),
                         pkey=ssh_info.get("pkey"))

    # ------------------------------ 读取 Kafka 相关配置 ------------------------------

    def kafka_list(self) -> List[str]:
        """获取 Kafka 列表"""
        return self._get_name_list("Kafka")

    def kafka_dict(self, name: str) -> Dict[str, Any]:
        return self._confirm_params(
            section="Kafka",
            config=self._get_section("Kafka", name),
            params=["bootstrap_servers"]
        )

    def kafka_obj(self, name: str) -> KafkaServer:
        """获取 Kafka Servers 对象"""
        kafka_info = self.kafka_dict(name)
        ssh_tunnel = self.ssh_tunnel_obj(kafka_info["use_ssh"]) if kafka_info.get("use_ssh") else None
        return KafkaServer(bootstrap_servers=kafka_info["bootstrap_servers"],
                           ssh_tunnel=ssh_tunnel)

    # ------------------------------ 读取 Kafka_Topic 相关配置 ------------------------------
    def kafka_topic_list(self) -> List[str]:
        """读取 Kafka-TOPIC 的配置列表"""
        return self._get_name_list("Kafka_Topic")

    def kafka_topic_dict(self, kafka_topic_name: str) -> Dict[str, Any]:
        """获取 Kafka-TOPIC 配置项的信息"""
        return self._confirm_params(
            section="Kafka_Topic",
            config=self._get_section("Kafka_Topic", kafka_topic_name),
            params=["kafka_server", "topic"]
        )

    def kafka_topic_obj(self, kafka_topic_name: str) -> KafkaTopic:
        """根据 Kafka-TOPIC 配置项的信息，构造 KafkaTopic 对象"""
        kafka_topic_info = self.kafka_topic_dict(kafka_topic_name)
        kafka_server = self.kafka_obj(kafka_topic_info["kafka_server"])
        return KafkaTopic(
            kafka_server=kafka_server,
            topic=kafka_topic_info["topic"]
        )

    # ------------------------------ 读取 Kafka_Topics_Group 相关配置 ------------------------------
    def kafka_topics_group_list(self) -> List[str]:
        """读取 Kafka_Topics_Group 的配置列表"""
        return self._get_name_list("Kafka_Topics_Group")

    def kafka_topics_group_dict(self, kafka_topics_group_name: str) -> Dict[str, Any]:
        """获取 Kafka_Topics_Group 配置项的信息"""
        return self._confirm_params(
            section="Kafka_Topics_Group",
            config=self._get_section("Kafka_Topics_Group", kafka_topics_group_name),
            params=["kafka_server", "topics", "group_id"]
        )

    def kafka_topics_group_obj(self, kafka_topics_group_name: str) -> KafkaTopicsGroup:
        """根据 Kafka-TOPIC 配置项的信息，构造 KafkaTopic 对象"""
        kafka_topic_group_info = self.kafka_topics_group_dict(kafka_topics_group_name)
        kafka_server = self.kafka_obj(kafka_topic_group_info["kafka_server"])
        return KafkaTopicsGroup(
            kafka_server=kafka_server,
            topics=kafka_topic_group_info["topics"],
            group_id=kafka_topic_group_info["group_id"]
        )

    def kafka_topics_group_new_consumer(self, kafka_topics_group_name: str) -> ConnKafkaConsumer:
        """根据 Kafka_Topic_Group 的名称，启动 KafkaConsumer 连接器"""
        kafka_topic_group = self.kafka_topics_group_obj(kafka_topics_group_name)
        return ConnKafkaConsumer.by_kafka_topic_group(kafka_topic_group)

    # ------------------------------ 读取 Hive 相关配置 ------------------------------

    def hive_list(self) -> List[str]:
        """获取 Hive 列表"""
        return self._get_name_list("Hive")

    def hive_dict(self, name: str) -> Dict[str, Any]:
        return self._confirm_params("Hive", self._get_section("Hive", name),
                                    ["hosts", "port"])

    def hive_obj(self, name: str) -> HiveInstance:
        """获取 Hive 列表"""
        hive_info = self.hive_dict(name)
        ssh_tunnel = self.ssh_tunnel_obj(hive_info["use_ssh"]) if hive_info.get("use_ssh") else None
        return HiveInstance(hosts=hive_info["hosts"], port=hive_info["port"], username=hive_info.get("username"),
                            ssh_tunnel=ssh_tunnel)

    # ------------------------------ 读取 DolphinScheduler 相关配置 ------------------------------

    def dolphin_meta_list(self) -> List[str]:
        """获取海豚调度元数据清单"""
        return self._get_name_list("DolphinMeta")

    def dolphin_meta_dict(self, name: str) -> Dict[str, Any]:
        """获取海豚调度元数据信息"""
        return self._confirm_params("DolphinMeta", self._get_section("DolphinMeta", name),
                                    ["host", "port", "user", "passwd"])

    def dolphin_meta_obj(self, name: str) -> DSMetaInstance:
        """获取海豚调度元数据的 DolphinMetaInstance 对象"""
        dolphin_meta_info = self.dolphin_meta_dict(name)
        ssh_tunnel = self.ssh_tunnel_obj(dolphin_meta_info["use_ssh"]) if dolphin_meta_info.get("use_ssh") else None
        return DSMetaInstance(
            host=dolphin_meta_info["host"],
            port=dolphin_meta_info["port"],
            user=dolphin_meta_info["user"],
            passwd=dolphin_meta_info["passwd"],
            db=dolphin_meta_info["db"],
            ssh_tunnel=ssh_tunnel
        )

    def dolphin_meta_new_connect(self, name: str):
        """根据海豚元数据名称，创建海豚元数据连接器"""
        dolphin_meta_instance = self.dolphin_meta_obj(name)
        return DSMetaConnector(dolphin_scheduler_meta_info=dolphin_meta_instance)

    # ------------------------------ 读取 OTS 相关配置 ------------------------------

    def ots_list(self) -> List[str]:
        """获取 OTS 元数据清单"""
        return self._get_name_list("OTS")

    def ots_dict(self, name: str) -> Dict[str, Any]:
        """获取 OTS 元数据信息"""
        return self._confirm_params("OTS", self._get_section("OTS", name),
                                    ["end_point", "access_key_id", "access_key_secret", "instance_name"])

    def ots_obj(self, name: str) -> OTSInstance:
        """获取 OTS 元数据的 OTSInstance 对象"""
        ots_info = self.ots_dict(name)
        return OTSInstance(
            end_point=ots_info["end_point"],
            access_key_id=ots_info["access_key_id"],
            access_key_secret=ots_info["access_key_secret"],
            instance_name=ots_info["instance_name"]
        )

    def otssql_connect(self, ots_name: str, use_pool: bool = True, **params):
        """根据 OTS 配置项中名称，获取连接（当 use_pool 为 True 时，从连接池中获取；否则，创建新连接）

        TODO 待支持实际的连接池
        """
        if use_pool is False:
            return OTSConnector(ots_instance=self.ots_obj(ots_name))
        if ots_name not in self._ots_conn_dict:
            self._ots_conn_dict[ots_name] = OTSConnector(ots_instance=self.ots_obj(ots_name), **params)
        return self._ots_conn_dict[ots_name]

    def otssql_new_connect(self, ots_name: str, **params):
        """根据实例名称，创建 MySQL 连接器"""
        return OTSConnector(ots_instance=self.ots_obj(ots_name), **params)

    # ------------------------------ 读取 Redis 相关配置 ------------------------------

    def redis_list(self) -> List[str]:
        """获取 Redis 元数据清单"""
        return self._get_name_list("Redis")

    def redis_dict(self, name: str) -> Dict[str, Any]:
        """获取 Redis 实例信息"""
        return self._confirm_params("Redis", self._get_section("Redis", name),
                                    ["host", "passwd", "port"])

    def redict_obj(self, name: str) -> RedisInstance:
        """获取 RedisInstance 对象"""
        redis_info = self.redis_dict(name)
        ssh_tunnel = self.ssh_tunnel_obj(redis_info["use_ssh"]) if redis_info.get("use_ssh") else None
        return RedisInstance(
            host=redis_info["host"],
            passwd=redis_info["passwd"],
            port=redis_info["port"],
            ssh_tunnel=ssh_tunnel
        )

    def redis_new_connect(self, name: str, db: int) -> RedisConnector:
        """根据 Redis 实例名称和数据库创建 Redis 连接"""
        redis_instance = self.redict_obj(name)
        return RedisConnector(RedisDatabase(instance=redis_instance, db=db))

    # ------------------------------ 读取 Redis_Database 相关配置 ------------------------------

    def redis_database_list(self):
        """获取 Redis_Database 清单"""
        return self._get_name_list("Redis_Database")

    def redis_database_dict(self, redis_database_name: str):
        """根据 Redis_Database 的配置项名称，获取 Redis_Database 信息"""
        return self._confirm_params(
            section="Redis_Database",
            config=self._get_section("Redis_Database", redis_database_name),
            params=["instance", "db"]
        )

    def redis_database_obj(self, redis_database_name: str) -> RedisDatabase:
        """根据 Redis_Database 的配置项名称，构造 RedisDatabase 对象"""
        redis_database_info = self.redis_database_dict(redis_database_name)
        redis_instance = self.redict_obj(redis_database_info["instance"])
        return RedisDatabase(instance=redis_instance, db=redis_database_info["db"])

    def redis_database_start_pool(self, redis_database_name: str) -> RedisConnectionPool:
        """根据 Redis_Database 的配置项名称，创建连接池

        Parameters
        ----------
        redis_database_name : str
            Redis_Database 配置项的名称

        Returns
        -------
        RedisConnectionPool
            Redis 连接池对象
        """
        if redis_database_name in self._redis_pool_dict:
            raise KeyError(f"重复启动 Redis_Database 连接池: {redis_database_name}")
        redis_database_obj = self.redis_database_obj(redis_database_name)
        self._redis_pool_dict[redis_database_name] = RedisConnectionPool(redis_database_obj)
        return self._redis_pool_dict[redis_database_name]

    def redis_database_close_pool(self, redis_database_name: str) -> None:
        """根据 Redis_Database 的配置项名称，关闭连接池"""
        if redis_database_name not in self._redis_pool_dict:
            raise KeyError(f"要关闭的 Redis_Database 连接池不存在: {redis_database_name}")
        self._redis_pool_dict[redis_database_name].close()  # 关闭连接池
        del self._redis_pool_dict[redis_database_name]  # 从缓存连接池中移除连接池

    def redis_database_connect(self, redis_database_name: str, use_pool: bool = True):
        """根据 Redis_Database 配置项中名称，获取连接（当 use_pool 为 True 时，从连接池中获取；否则，创建新连接）"""
        if use_pool is True:
            if redis_database_name not in self._redis_pool_dict:
                self.redis_database_start_pool(redis_database_name)
            connection_pool = self._redis_pool_dict[redis_database_name]
            connection_pool = typing.cast(redis.ConnectionPool, connection_pool)
            return redis.Redis(connection_pool=connection_pool)
        return RedisConnector(self.redis_database_obj(redis_database_name))

    def redis_database_new_connect(self, name: str):
        """根据 Redis_Database 配置项中的名称创建 Redis 连接"""
        return RedisConnector(self.redis_database_obj(name))

    # ---------- 其他工具方法 ----------

    def _get_name_list(self, section: str):
        """获取每种配置项中的所有名称的列表"""
        if section not in self._configuration:
            return []
        return list(self._configuration[section])

    def _get_section(self, section: str, name: str) -> Dict[str, Any]:
        """获取每种类型的配置信息数据"""
        if section not in self._configuration or name not in self._configuration[section]:
            return {}
        config = self._configuration[section][name].get("mode:common", {}).copy()  # 先加载通用配置
        config.update(self._configuration[section][name].get(f"mode:{self._mode}", {}))  # 然后再加对应模式的配置
        return config

    @staticmethod
    def _confirm_params(section: str, config: Dict[str, Any], params: List[str]) -> Dict[str, Any]:
        """检查参数是否满足"""
        for param in params:
            assert param in config, f"param {param} not in {section} config"
        return config

    @classmethod
    def make_templates(cls, path: str):
        """在指定路径下生成模板配置文件"""
        if os.path.exists(path):
            raise ValueError(f"配置文件已存在({path}),创建模板文件失败!")
        with open(path, "w", encoding=cls.ENCODING) as file:
            file.write(json.dumps(
                {"MySQL": {"localhost": {"host": "localhost", "port": 3306, "user": "root", "passwd": "123456"}},
                 "SSH": {"demo": {"host": "...", "port": "...", "username": "...", "pkey": "..."}}}))
            print("配置文件模板生成完成")

    def close(self, is_del: bool = False):
        """关闭连接管理器中的所有连接、连接池

        Parameters
        ----------
        is_del : bool, default = False
            是否是在对象销毁的过程中调用：如果是在对象销毁的过程中调用，则不销毁 SshTunnel 对象，因为该对象在销毁过程中调用 close 方法可能出现
            无法预料的报错或卡死
        """
        # 关闭 Redis 连接池
        for redis_connection_pool in self._redis_pool_dict.values():
            redis_connection_pool.close(is_del=True)

        # 关闭 MySQL 连接池
        for mysql_connection_pool in self._mysql_pool_dict.values():
            mysql_connection_pool.close(is_del=True)

    def __del__(self):
        """在对象被销毁时调用"""
        self.close(is_del=True)



"""
连接器通用工具类
"""

__all__ = [
    "WrapConnector"
]


class WrapConnector:
    """嵌套连接器（以提供 with 开关的功能）"""

    def __init__(self, connection):
        self._mysql_con = connection
        self._is_close = False

    def close(self):
        if self._is_close is False:
            if self._mysql_con is not None:
                self._mysql_con.close()
            self._is_close = True

    def __getattr__(self, name):
        """代理嵌套连接类中的所有成员"""
        return getattr(self._mysql_con, name)

    def __enter__(self):
        """在进入 WITH 语句时被调用（返回值作为 AS 后面的变量）"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """在离开 WITH 语句时被调用"""
        self.close()

    def __del__(self):
        """在对象被销毁时调用"""
        self.close()



"""
通用工具类
"""

import enum
from typing import Any, List, Optional

__all__ = [
    "to_quote_str_none_as_null",
    "to_quote_str_none_as_zero_str",
    "to_quote_str_none_as_empty_str"
]


class NoneAs(enum.Enum):
    """处理 None 的方法"""

    ASSERT = None  # 不允许为空
    IGNORE = None  # 跳过 None 值
    NULL = "null"  # 将 None 视作 null
    EMPTY_STR = "''"  # 将 None 视作空字符串
    ZERO_STR = "'0'"  # 将 None 视作包含 0 的字符串

    @property
    def is_visible(self) -> bool:
        """如果当前配置下 None 值是否可以转化为字面值则返回 True，否则返回 False"""
        return self.value is not None


def to_quote_str(text: Optional[Any], none_as: NoneAs) -> str:
    """将 text 转化为 sql 的单引号字符串

    Parameters
    ----------
    text : Optional[str]
        需要转化为 sql 的单引号字符串的值
    none_as : NoneAs
        处理 None 值的方法

    Returns
    -------
    str
        sql 的单引号字符串
    """
    if text is None:
        if not none_as.is_visible:
            raise KeyError(f"none_as={none_as} 时 None 值无法被直接转化为 SQL 字面值")
        return none_as.value
    if isinstance(text, str):
        text = text.replace("\\", "\\\\").replace("'", "\\'")
    return f"'{text}'"


def to_quote_str_none_as_null(text: Optional[Any]) -> str:
    """将 text 转化为 sql 的单引号字符串并将 None 值处理为 null"""
    return to_quote_str(text, none_as=NoneAs.NULL)


def to_quote_str_none_as_empty_str(text: Optional[Any]) -> str:
    """将 text 转化为 sql 的单引号字符串并将 None 值处理为空字符串"""
    return to_quote_str(text, none_as=NoneAs.EMPTY_STR)


def to_quote_str_none_as_zero_str(text: Optional[Any]) -> str:
    """将 text 转化为 sql 的单引号字符串并将 None 值处理为 '0' 字符串"""
    return to_quote_str(text, none_as=NoneAs.ZERO_STR)


def to_quote_str_list(text_list: List[Optional[Any]], none_as: NoneAs) -> str:
    """将 text_list 转化用括号包含的单引号字符串的列表"""
    if none_as == NoneAs.IGNORE:
        quote_str_list = [to_quote_str(text, none_as=NoneAs.ASSERT) for text in text_list if text is not None]
    else:
        quote_str_list = [to_quote_str(text, none_as=none_as) for text in text_list]
    return "(" + ",".join(quote_str_list) + ")"


def to_quote_str_list_none_as_ignore(text_list: List[Optional[Any]]) -> str:
    """将 text_list 转化用括号包含的单引号字符串的列表，并将 None 值忽略"""
    return to_quote_str_list(text_list, none_as=NoneAs.IGNORE)


def to_quote_str_list_none_as_null(text_list: List[Optional[Any]]) -> str:
    """将 text_list 转化用括号包含的单引号字符串的列表，并将 None 值则置为 null"""
    return to_quote_str_list(text_list, none_as=NoneAs.NULL)



import dataclasses
from typing import Optional

from metasequoia_connector.node.ssh_tunnel import SshTunnel

__all__ = ["MysqlInstance", "MysqlTable"]


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class MysqlInstance:
    """MySQL 实例"""

    host: str = dataclasses.field(kw_only=True)
    port: int = dataclasses.field(kw_only=True)
    user: str = dataclasses.field(kw_only=True)
    passwd: str = dataclasses.field(kw_only=True)
    ssh_tunnel: Optional[SshTunnel] = dataclasses.field(kw_only=True, default=None)


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class MysqlTable:
    """MySQL 表"""

    instance: MysqlInstance = dataclasses.field(kw_only=True)
    schema: str = dataclasses.field(kw_only=True)
    table: str = dataclasses.field(kw_only=True)



"""
MySQL 连接器
"""

from typing import Optional

import dbutils
import pymysql
import pymysql.cursors

from metasequoia_connector.connector.ssh_tunnel import create_ssh_tunnel
from metasequoia_connector.node import MysqlInstance

__all__ = [
    "MysqlConnectionPool",
    "MysqlConnector",
]


class MysqlConnectionPool:
    """MySQL 连接池"""

    def __init__(self,
                 mysql_instance: MysqlInstance,
                 schema: Optional[str] = None,
                 connect_timeout: int = 5,
                 read_timeout: int = 10) -> None:
        """MySQL 连接的构造方法

        Parameters
        ----------
        mysql_instance : MysqlInstance
            MySQL 实例对象
        schema : Optional[str], default = None
            数据库名称
        connect_timeout : int, default = 10
            连接超时时间
        read_timeout : int, default = 30
            读取超时时间
        """
        self._ssh_tunnel_con = None
        self._is_close = False

        # 初始化 MySQL 连接和 SSH 隧道连接
        if mysql_instance.ssh_tunnel is not None:
            # 启动 SSH 隧道
            self._ssh_tunnel_con = create_ssh_tunnel(
                mysql_instance.ssh_tunnel,
                remote_bind_address=(mysql_instance.host, mysql_instance.port)
            )

            # 更新 MySQL 连接信息，令 MySQL 连接到 SSH 隧道
            host = "127.0.0.1"
            port = self._ssh_tunnel_con.local_bind_port
        else:
            host = mysql_instance.host
            port = mysql_instance.port

        # 启动 MySQL 连接
        self._mysql_pool = dbutils.pooled_db.PooledDB(
            creator=pymysql,
            host=host,
            port=port,
            user=mysql_instance.user,
            passwd=mysql_instance.passwd,
            db=schema,
            connect_timeout=connect_timeout,
            read_timeout=read_timeout
        )

    def close(self, is_del: bool = False):
        if self._is_close is False:
            if self._mysql_pool is not None:
                # noinspection PyBroadException
                try:
                    self._mysql_pool.close()
                except Exception:
                    pass
            if self._ssh_tunnel_con is not None and is_del is False:
                # 如果是在对象销毁的过程中调用，则不销毁 SshTunnel 对象，因为该对象在销毁过程中调用 close 方法可能出现无法预料的报错或卡死
                # noinspection PyBroadException
                try:
                    self._ssh_tunnel_con.stop()
                except Exception:
                    pass
            self._is_close = True

    def __getattr__(self, name):
        """代理嵌套连接类中的所有成员"""
        return getattr(self._mysql_pool, name)

    def __enter__(self):
        """在进入 WITH 语句时被调用（返回值作为 AS 后面的变量）"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """在离开 WITH 语句时被调用"""
        self.close()

    def __del__(self):
        """在对象被销毁时调用"""
        self.close(is_del=True)


class MysqlConnector:
    def __init__(self,
                 mysql_instance: MysqlInstance,
                 schema: Optional[str] = None,
                 connect_timeout: int = 5,
                 read_timeout: int = 10) -> None:
        """MySQL 连接的构造方法

        Parameters
        ----------
        mysql_instance : MysqlInstance
            MySQL 实例对象
        schema : Optional[str], default = None
            数据库名称
        connect_timeout : int, default = 10
            连接超时时间
        read_timeout : int, default = 30
            读取超时时间
        """
        self._ssh_tunnel_con = None
        self._is_close = False

        # 初始化 MySQL 连接和 SSH 隧道连接
        if mysql_instance.ssh_tunnel is not None:
            # 启动 SSH 隧道
            self._ssh_tunnel_con = create_ssh_tunnel(
                mysql_instance.ssh_tunnel,
                remote_bind_address=(mysql_instance.host, mysql_instance.port)
            )

            # 更新 MySQL 连接信息，令 MySQL 连接到 SSH 隧道
            host = "127.0.0.1"
            port = self._ssh_tunnel_con.local_bind_port
        else:
            host = mysql_instance.host
            port = mysql_instance.port

        # 启动 MySQL 连接
        self._mysql_con = pymysql.connect(
            host=host,
            port=port,
            user=mysql_instance.user,
            passwd=mysql_instance.passwd,
            db=schema,
            connect_timeout=connect_timeout,
            read_timeout=read_timeout
        )

    def close(self, is_del: bool = False):
        if self._is_close is False:
            if self._mysql_con is not None:
                # noinspection PyBroadException
                try:
                    self._mysql_con.close()
                except Exception:
                    pass
            if self._ssh_tunnel_con is not None and is_del is False:
                # 如果是在对象销毁的过程中调用，则不销毁 SshTunnel 对象，因为该对象在销毁过程中调用 close 方法可能出现无法预料的报错或卡死
                # noinspection PyBroadException
                try:
                    self._ssh_tunnel_con.stop()
                except Exception:
                    pass
            self._is_close = True

    def __getattr__(self, name):
        """代理嵌套连接类中的所有成员"""
        return getattr(self._mysql_con, name)

    def __enter__(self):
        """在进入 WITH 语句时被调用（返回值作为 AS 后面的变量）"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """在离开 WITH 语句时被调用"""
        self.close()

    def __del__(self):
        """在对象被销毁时调用"""
        self.close(is_del=True)



"""
MySQL 相关工具类
"""

import dataclasses
from typing import Any, Dict, Generator, List, Optional, Tuple

import pymysql
import pymysql.cursors

import metasequoia_sql as ms_sql
from metasequoia_connector.connector import MysqlConnector
from metasequoia_connector.manager.connect_manager import ConnectManager
from metasequoia_connector.node import MysqlInstance, SshTunnel
from metasequoia_connector.utils.sql_format import to_quote_str_none_as_null


# ---------------------------------------- 查询方法 ----------------------------------------


def conn_select_all_as_dict(conn: pymysql.Connection, sql: str) -> Tuple[Dict[str, Any], ...]:
    """【查询】以字典列表格式返回所有结果"""
    with conn.cursor(pymysql.cursors.DictCursor) as cursor:
        cursor.execute(sql)
        return cursor.fetchall()


def select_all_as_dict(manager: ConnectManager, mysql_name: str, db_name: str, sql: str):
    """【查询】以字典列表格式返回所有结果"""
    with manager.mysql_connect(mysql_name, db_name) as mysql_conn:
        return conn_select_all_as_dict(conn=mysql_conn, sql=sql)


def conn_select_one_as_dict(conn: pymysql.Connection, sql: str) -> Dict[str, Any]:
    """【查询】以字典格式返回一条结果"""
    with conn.cursor(pymysql.cursors.DictCursor) as cursor:
        cursor.execute(sql)
        return cursor.fetchone()


def select_one_as_dict(manager: ConnectManager, mysql_name: str, db_name: str, sql: str):
    """【查询】以字典格式返回一条结果"""
    with manager.mysql_connect(mysql_name, db_name) as mysql_conn:
        return conn_select_one_as_dict(conn=mysql_conn, sql=sql)


def conn_execute_and_commit(conn: pymysql.Connection, sql: str) -> int:
    """【执行】执行 SQL 语句并提交，返回影响行数"""
    with conn.cursor(pymysql.cursors.DictCursor) as cursor:
        result = cursor.execute(sql)
    conn.commit()
    return result


def execute_and_commit(manager: ConnectManager, mysql_name: str, db_name: str, sql: str) -> int:
    """【执行】执行 SQL 语句并提交，返回影响行数"""
    with manager.mysql_connect(mysql_name, db_name) as mysql_conn:
        return conn_execute_and_commit(conn=mysql_conn, sql=sql)


def conn_execute_and_commit_with_args(conn: pymysql.Connection, sql: str, *args) -> int:
    with conn.cursor() as cursor:
        result = cursor.execute(sql, args)
    conn.commit()
    return result


def show_databases(rds_instance: MysqlInstance):
    """执行：SHOW DATABASES"""
    with MysqlConnector(mysql_instance=rds_instance) as conn:
        return conn_show_databases(conn)


def show_tables(rds_instance: MysqlInstance, schema: str):
    """执行：SHOW TABLES"""
    with MysqlConnector(mysql_instance=rds_instance, schema=schema) as conn:
        return conn_show_tables(conn)


def show_create_table(rds_instance: MysqlInstance, schema: str, table: str,
                      ssh_tunnel: Optional[SshTunnel] = None):
    """执行：SHOW CREATE TABLE"""
    with MysqlConnector(mysql_instance=rds_instance, schema=schema) as conn:
        return conn_show_create_table(conn, table)


def conn_use(conn: pymysql.Connection, schema: str) -> None:
    with conn.cursor() as cursor:
        cursor.execute(f"USE `{schema}`")


def conn_select_all(conn: pymysql.Connection, sql: str) -> Tuple[Tuple[Any, ...], ...]:
    with conn.cursor() as cursor:
        cursor.execute(sql)
        return cursor.fetchall()


def select_iter_as_dict(manager: ConnectManager, mysql_name: str, db_name: str, sql: str,
                        primary_key: str, batch_size: int = 1000) -> Generator[Dict[str, Any], None, None]:
    """【执行】执行 SQL 语句并提交，返回影响行数"""
    with manager.mysql_connect(mysql_name, db_name) as mysql_conn:
        yield from conn_select_iter_as_dict(conn=mysql_conn, sql=sql, primary_key=primary_key, batch_size=batch_size)


def conn_select_iter_as_dict(conn: pymysql.Connection, sql: str, primary_key: str, batch_size: int = 1000
                             ) -> Generator[Dict[str, Any], None, None]:
    """使用 sql 查询 conn 连接的 MySQL，并将结果作为 dict 的迭代器返回

    使用基于 primary_key 的深翻页方法，要求：

    - primary_key 字段为主键或为唯一键（有索引，且在 WHERE 条件后仍然有索引可用于排序）
    - 提供的 SQL 语句中不能包含 GROUP BY 子句、ORDER BY 子句和 LIMIT 子句
    - 提供的 SQL 语句中查询的字段需要包含主键字段

    Parameters
    ----------
    conn : pymysql.Connection
        Mysql 连接
    sql : str
        SQL 语句
    primary_key : str
        主键或唯一键
    batch_size : int, default = 1000
        每次翻页多少条记录

    Yields
    ------
    Dict[str, Any]
        查询结果记录
    """
    # 解析并检查 SQL 语句
    select_statement: ms_sql.ASTSingleSelectStatement = ms_sql.SQLParser.parse_single_select_statement(sql)
    if select_statement.group_by_clause is not None:
        raise KeyError(f"SQL 语句包含 ORDER BY 子句：{sql}")
    if select_statement.order_by_clause is not None:
        raise KeyError(f"SQL 语句包含 ORDER BY 子句：{sql}")
    if select_statement.limit_clause is not None:
        raise KeyError(f"SQL 语句包含 LIMIT 子句：{sql}")

    with conn.cursor(pymysql.cursors.DictCursor) as cursor:
        # 构造第一次请求的语句
        actual_sql = ms_sql.ASTSingleSelectStatement(
            with_clause=select_statement.with_clause,
            select_clause=select_statement.select_clause,
            from_clause=select_statement.from_clause,
            lateral_view_clauses=select_statement.lateral_view_clauses,
            join_clauses=select_statement.join_clauses,
            where_clause=select_statement.where_clause,
            group_by_clause=select_statement.group_by_clause,
            having_clause=select_statement.having_clause,
            order_by_clause=ms_sql.SQLParser.parse_order_by_clause(f"ORDER BY {primary_key}"),
            limit_clause=ms_sql.ASTLimitClause(limit=batch_size, offset=0)
        ).source(ms_sql.SQLType.MYSQL)

        max_primary_key = None
        while True:
            n_fetch = cursor.execute(actual_sql)

            for query_row in cursor.fetchall():
                max_primary_key = query_row[primary_key]  # 更新当前主键最大值（用于下一次请求）
                yield query_row

            if n_fetch < batch_size:
                break  # 如果当前页数量少于 LIMIT 数量，则说明已经是最后一页，跳出循环

            if select_statement.where_clause is None:
                # 初始原始 SQL 语句没有 WHERE 子句的情况
                actual_sql = ms_sql.ASTSingleSelectStatement(
                    with_clause=select_statement.with_clause,
                    select_clause=select_statement.select_clause,
                    from_clause=select_statement.from_clause,
                    lateral_view_clauses=select_statement.lateral_view_clauses,
                    join_clauses=select_statement.join_clauses,
                    where_clause=ms_sql.SQLParser.parse_where_clause(f"WHERE {primary_key} > '{max_primary_key}'"),
                    group_by_clause=select_statement.group_by_clause,
                    having_clause=select_statement.having_clause,
                    order_by_clause=ms_sql.SQLParser.parse_order_by_clause(f"ORDER BY {primary_key}"),
                    limit_clause=ms_sql.ASTLimitClause(limit=batch_size, offset=0)
                ).source(ms_sql.SQLType.MYSQL)
            else:
                # 初始原始 SQL 语句有 WHERE 子句的情况
                where_condition_str = select_statement.where_clause.condition.source(ms_sql.SQLType.MYSQL)
                actual_sql = ms_sql.ASTSingleSelectStatement(
                    with_clause=select_statement.with_clause,
                    select_clause=select_statement.select_clause,
                    from_clause=select_statement.from_clause,
                    lateral_view_clauses=select_statement.lateral_view_clauses,
                    join_clauses=select_statement.join_clauses,
                    where_clause=ms_sql.SQLParser.parse_where_clause(
                        f"WHERE ({where_condition_str}) AND {primary_key} > '{max_primary_key}'"),
                    group_by_clause=select_statement.group_by_clause,
                    having_clause=select_statement.having_clause,
                    order_by_clause=ms_sql.SQLParser.parse_order_by_clause(f"ORDER BY {primary_key}"),
                    limit_clause=ms_sql.ASTLimitClause(limit=batch_size, offset=0)
                ).source(ms_sql.SQLType.MYSQL)


def conn_execute_multi_and_commit(conn: pymysql.Connection, sql_list: List[str]) -> int:
    """在同一个事务中执行多个 SQL 语句，并 commit"""
    with conn.cursor(pymysql.cursors.DictCursor) as cursor:
        for sql in sql_list:
            result = cursor.execute(sql)
        conn.commit()
        return result


def conn_show_databases(conn: pymysql.Connection) -> List[str]:
    """执行 SHOW DATABASES 语句，返回数据库名称的列表"""
    return [row["Database"] for row in conn_select_all_as_dict(conn, "SHOW DATABASES")]


def conn_show_tables(conn: pymysql.Connection) -> List[str]:
    """执行 SHOW TABLES 语句，返回表名的列表"""
    return [row[0] for row in conn_select_all(conn, "SHOW TABLES")]


def conn_show_create_table(conn: pymysql.Connection, table: str) -> Optional[str]:
    """执行 SHOW CREATE TABLE 语句

    无法获取时返回 None，在以下场景下无法获取：
    1. 账号没有权限
    2. 表名不存在

    实现说明：
    1. 为使名称完全为数字的表执行正常，所以在表名外添加了引号
    """
    result = conn_select_all_as_dict(conn, f"SHOW CREATE TABLE `{table}`")[0]
    if "Create Table" in result:
        return result["Create Table"]
    else:  # 当表名不存在或没有权限时，没有 Create Table 列，只有 Error 列
        return None


def conn_insert_into_by_dict_list(conn: pymysql.Connection,
                                  table_name: str,
                                  write_list: List[dict],
                                  batch_size: int = 1000,
                                  debug: bool = False) -> int:
    """将 data 中的 dataclass 格式数据写入到 table_name 中

    Parameters
    ----------
    conn : pymysql.Connection
        MySQL 连接
    table_name : str
        表名，可以是 table_name 或 schema_name.table_name
    write_list : List[dataclasses.dataclass]
        记录的列表，要求每条记录拥有相同的结构（将根据第 1 条记录获取数据结构）
    batch_size : int, default = 1000
        每一批的数量
    debug : bool, default = False
        是否开启调试模式（开启时打印进度日志）

    Returns
    -------
    int
        影响的记录数
    """
    if len(write_list) == 0:
        return 0  # 如果没有要写入记录则返回 0

    # 获取所有列的清单（遍历所有记录获取全集）
    column_set = set()
    for item in write_list:
        column_set |= set(item.keys())
    column_list = list(column_set)
    column_list_str = ",".join(column_list)

    # 获取每条记录的值的列表
    result = 0
    value_list = []
    for i, write_data in enumerate(write_list):
        # 获取当前记录中每个字段的格式化后的值的列表
        cell_list = [to_quote_str_none_as_null(write_data.get(column)) for column in column_list]
        value_str = ",".join(cell_list)
        value_list.append(f"({value_str})")
        if len(value_list) >= batch_size:
            value_list_str = ",".join(value_list)
            sql = f"INSERT INTO `{table_name}` ({column_list_str}) VALUES {value_list_str}"
            result += conn_execute_and_commit(conn, sql)
            value_list = []
            if debug is True:
                print(f"already write {i} / {len(write_list)}")
    if value_list:
        value_list_str = ",".join(value_list)
        sql = f"INSERT INTO `{table_name}` ({column_list_str}) VALUES {value_list_str}"
        result += conn_execute_and_commit(conn, sql)
    return result


def conn_insert_dataclass_list(conn: pymysql.Connection,
                               table_name: str,
                               write_list: List[dataclasses.dataclass]) -> int:
    """将 data 中的 dataclass 格式数据写入到 table_name 中

    Parameters
    ----------
    conn : pymysql.Connection
        MySQL 连接
    table_name : str
        表名，可以是 table_name 或 schema_name.table_name
    write_list : List[dataclasses.dataclass]
        记录的列表，要求每条记录拥有相同的结构（将根据第 1 条记录获取数据结构）

    Returns
    -------
    int
        影响的记录数
    """
    if len(write_list) == 0:
        return 0  # 如果没有要写入记录则返回 0

    # 获取所有列的清单（按第 1 条记录）
    column_list = [field.name for field in dataclasses.fields(write_list[0])]
    column_list_str = ",".join(f"`{column_name}`" for column_name in column_list)

    # 获取每条记录的值的列表
    value_list = []
    for write_data in write_list:
        # 获取当前记录中每个字段的格式化后的值的列表
        cell_list = [to_quote_str_none_as_null(getattr(write_data, column))
                     for column in column_list]
        value_str = ",".join(cell_list)
        value_list.append(f"({value_str})")
    value_list_str = ",".join(value_list)
    sql = f"INSERT INTO `{table_name}` ({column_list_str}) VALUES {value_list_str}"
    return conn_execute_and_commit(conn, sql)


def insert_dataclass_list(manager: ConnectManager, mysql_name: str, db_name: str,
                          table_name: str,
                          write_list: List[dataclasses.dataclass]) -> int:
    """【执行】执行 SQL 语句并提交，返回影响行数"""
    with manager.mysql_connect(mysql_name, db_name) as mysql_conn:
        return conn_insert_dataclass_list(conn=mysql_conn, table_name=table_name, write_list=write_list)


def conn_insert_or_update_dataclass(conn: pymysql.Connection,
                                    table_name: str,
                                    write_data: dataclasses.dataclass) -> int:
    """将 dataclass 对象写入到 table_name 中

    Parameters
    ----------
    conn : pymysql.Connection
        MySQL 连接
    table_name : str
        表名，可以是 table_name 或 schema_name.table_name
    write_data : dataclasses.dataclass
        dataclasses 对象

    Returns
    -------
    int
        影响的记录数
    """
    # 获取 dataclasses 对象中的所有列
    column_name_list = [field.name for field in dataclasses.fields(write_data)]

    # 生成所有列的清单
    column_name_list_str = ",".join(f"`{column_name}`" for column_name in column_name_list)

    # 生成每个字段格式化后的值的列表
    value_list = [to_quote_str_none_as_null(getattr(write_data, column_name)) for column_name in column_name_list]
    value_list_str = ",".join(value_list)

    # 生成 ON DUPLICATE KEY UPDATE 语句
    update_list = [f"`{column_name}` = VALUES(`{column_name}`)" for column_name in column_name_list]
    update_list_str = ",".join(update_list)

    sql = (f"INSERT INTO `{table_name}` ({column_name_list_str}) "
           f"VALUES ({value_list_str}) "
           f"ON DUPLICATE KEY UPDATE {update_list_str}")

    return conn_execute_and_commit(conn, sql)


def insert_or_update_dataclass(manager: ConnectManager, mysql_name: str, db_name: str,
                               table_name: str,
                               write_data: dataclasses.dataclass) -> int:
    """【执行】执行 SQL 语句并提交，返回影响行数"""
    with manager.mysql_connect(mysql_name, db_name) as mysql_conn:
        return conn_insert_or_update_dataclass(conn=mysql_conn, table_name=table_name, write_data=write_data)


def conn_insert_into_or_update_by_dataclass_list(conn: pymysql.Connection,
                                                 table_name: str,
                                                 unique_key: List[str],
                                                 write_list: List[dataclasses.dataclass]) -> int:
    """将 data 中的 dataclass 格式数据写入到 table_name 中，如果有重复值，则更新除 unique_key 之外的字段

    语句逻辑：
    INSERT INTO ... ON DUPLICATE KEY UPDATE

    Parameters
    ----------
    conn : pymysql.Connection
        MySQL 连接
    table_name : str
        表名，可以是 table_name 或 schema_name.table_name
    unique_key : List[str]
        表唯一键，在 ON DUPLICATE KEY UPDATE 时不更新这些字段
    write_list : List[dataclasses.dataclass]
        记录的列表，要求每条记录拥有相同的结构（将根据第 1 条记录获取数据结构）

    Returns
    -------
    int
        影响的记录数
    """
    if len(write_list) == 0:
        return 0  # 如果没有要写入记录则返回 0

    # 获取所有列的清单（按第 1 条记录）
    column_list = [field.name for field in dataclasses.fields(write_list[0])]
    column_list_str = ",".join(column_list)

    # 计算所有需要更新的字段
    update_column_list = [f"`{field}`=VALUES(`{field}`)" for field in column_list if field not in unique_key]
    update_column_list_str = ",".join(update_column_list)

    # 获取每条记录的值的列表
    value_list = []
    for write_data in write_list:
        # 获取当前记录中每个字段的格式化后的值的列表
        cell_list = [to_quote_str_none_as_null(getattr(write_data, column))
                     for column in column_list]
        value_str = ",".join(cell_list)
        value_list.append(f"({value_str})")
    value_list_str = ",".join(value_list)
    sql = (f"INSERT INTO `{table_name}` ({column_list_str}) VALUES {value_list_str} "
           f"ON DUPLICATE KEY UPDATE {update_column_list_str}")
    return conn_execute_and_commit(conn, sql)


def insert_or_update_dataclass_list(manager: ConnectManager, mysql_name: str, db_name: str,
                                    unique_key: List[str],
                                    table_name: str,
                                    write_list: List[dataclasses.dataclass]) -> int:
    """【执行】执行 SQL 语句并提交，返回影响行数"""
    with manager.mysql_connect(mysql_name, db_name) as mysql_conn:
        return conn_insert_into_or_update_by_dataclass_list(conn=mysql_conn, table_name=table_name,
                                                            unique_key=unique_key, write_list=write_list)



"""
Redis 节点
"""

import dataclasses
from typing import Optional

from metasequoia_connector.node.ssh_tunnel import SshTunnel

__all__ = ["RedisInstance", "RedisDatabase"]


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class RedisInstance:
    """Redis 实例"""

    host: str = dataclasses.field(kw_only=True)
    passwd: str = dataclasses.field(kw_only=True)
    port: int = dataclasses.field(kw_only=True)
    ssh_tunnel: Optional[SshTunnel] = dataclasses.field(kw_only=True, default=None)


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class RedisDatabase:
    """Redis 数据库"""

    instance: RedisInstance = dataclasses.field(kw_only=True)
    db: int = dataclasses.field(kw_only=True)



"""
MySQL 连接器
"""

import redis

from metasequoia_connector.connector.ssh_tunnel import create_ssh_tunnel
from metasequoia_connector.node import RedisDatabase

__all__ = [
    "RedisConnector",
    "RedisConnectionPool",
]


class RedisConnectionPool:
    """Redis 连接池"""

    def __init__(self, redis_database: RedisDatabase) -> None:
        """Redis 连接的构造方法

        Parameters
        ----------
        redis_database : MysqlInstance
            Redis 实例对象
        """
        self._ssh_tunnel_con = None
        self._is_close = False

        if redis_database.instance.ssh_tunnel is not None:
            # 启动 SSH 隧道
            self._ssh_tunnel_con = create_ssh_tunnel(
                redis_database.instance.ssh_tunnel,
                remote_bind_address=(redis_database.instance.host, redis_database.instance.port)
            )

            # 更新 Redis 连接信息，令 Redis 连接到 SSH 隧道
            host = "127.0.0.1"
            port = self._ssh_tunnel_con.local_bind_port
        else:
            host = redis_database.instance.host
            port = redis_database.instance.port

        # 启动 Redis 连接池
        self._connection_pool = redis.ConnectionPool(
            host=host,
            password=redis_database.instance.passwd,
            db=redis_database.db,
            port=port
        )

    def close(self, is_del: bool = False):
        if self._is_close is False:
            if self._connection_pool is not None:
                # noinspection PyBroadException
                try:
                    self._connection_pool.disconnect()
                except Exception:
                    pass
            if self._ssh_tunnel_con is not None and is_del is False:
                # 如果是在对象销毁的过程中调用，则不销毁 SshTunnel 对象，因为该对象在销毁过程中调用 close 方法可能出现无法预料的报错或卡死
                # noinspection PyBroadException
                try:
                    self._ssh_tunnel_con.stop()
                except Exception:
                    pass
            self._is_close = True

    def __getattr__(self, name):
        """代理嵌套连接类中的所有成员"""
        return getattr(self._connection_pool, name)

    def __enter__(self):
        """在进入 WITH 语句时被调用（返回值作为 AS 后面的变量）"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """在离开 WITH 语句时被调用"""
        self.close()

    def __del__(self):
        """在对象被销毁时调用"""
        self.close(is_del=True)


class RedisConnector:
    def __init__(self, redis_database: RedisDatabase) -> None:
        """Redis 连接的构造方法

        Parameters
        ----------
        redis_database : MysqlInstance
            Redis 实例对象
        """
        self._ssh_tunnel_con = None
        self._is_close = False

        if redis_database.instance.ssh_tunnel is not None:
            # 启动 SSH 隧道
            self._ssh_tunnel_con = create_ssh_tunnel(
                redis_database.instance.ssh_tunnel,
                remote_bind_address=(redis_database.instance.host, redis_database.instance.port)
            )

            # 更新 Redis 连接信息，令 Redis 连接到 SSH 隧道
            host = "127.0.0.1"
            port = self._ssh_tunnel_con.local_bind_port
        else:
            host = redis_database.instance.host
            port = redis_database.instance.port

        # 启动 Redis 连接
        connection_pool = redis.ConnectionPool(
            host=host,
            password=redis_database.instance.passwd,
            db=redis_database.db,
            port=port
        )
        self._redis_con = redis.Redis(connection_pool=connection_pool)

    def close(self, is_del: bool = False):
        if self._is_close is False:
            if self._redis_con is not None:
                # noinspection PyBroadException
                try:
                    self._redis_con.close()
                except Exception:
                    pass
            if self._ssh_tunnel_con is not None and is_del is False:
                # 如果是在对象销毁的过程中调用，则不销毁 SshTunnel 对象，因为该对象在销毁过程中调用 close 方法可能出现无法预料的报错或卡死
                # noinspection PyBroadException
                try:
                    self._ssh_tunnel_con.stop()
                except Exception:
                    pass
            self._is_close = True

    def __getattr__(self, name):
        """代理嵌套连接类中的所有成员"""
        return getattr(self._redis_con, name)

    def __enter__(self):
        """在进入 WITH 语句时被调用（返回值作为 AS 后面的变量）"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """在离开 WITH 语句时被调用"""
        self.close()

    def __del__(self):
        """在对象被销毁时调用"""
        self.close(is_del=True)



from typing import Optional

from metasequoia_connector.connector import *
from metasequoia_connector.manager import ConnectManager
from metasequoia_connector.node import *
from metasequoia_connector.utils import dolphin, hive, kafka, mysql, otssql, sql_format


def from_environment(environ_name: Optional[str] = None, mode: str = "dev") -> ConnectManager:
    """从环境变量中读取配置文件路径，并加载配置文件"""
    return ConnectManager.from_environment(environ_name, mode=mode)


# 定义连接器的别名
connect_dolphin_meta = DSMetaConnector  # 定义海豚元数据连接器的别名
connect_hive = HiveConnector  # 定义 Hive 连接器的别名
connect_kafka_admin_client = ConnKafkaAdminClient  # 定义 Kafka 管理客户端连接器的别名
connect_kafka_consumer = ConnKafkaConsumer  # 定义 Kafka 消费者连接器的别名
connect_kafka_consumer_by_kafka_topics_group = ConnKafkaConsumer.by_kafka_topic_group  # 使用 KafkaTopicsGroup 构造 KafkaConsumer
connect_kafka_producer = ConnKafkaProducer  # 定义 Kafka 生产者连接器的别名
connect_mysql = MysqlConnector  # 定义 MySQL 连接器的别名
connect_ots = OTSConnector  # 定义 OTS 连接器的别名
create_ssh_tunnel = create_ssh_tunnel  # 定义 SSH 隧道连接器的别名

del Optional



"""
Kafka 工具函数
"""

from typing import Any, Dict, List

from kafka.admin import ConfigResource, ConfigResourceType
from kafka.protocol.admin import DescribeConfigsResponse_v2

from metasequoia_connector.connector import ConnKafkaAdminClient
from metasequoia_connector.node import KafkaGroup, KafkaServer, KafkaTopic


def list_topics(kafka_server: KafkaServer) -> List[str]:
    """列出 Kafka 集群中所有的 TOPIC"""
    with ConnKafkaAdminClient(kafka_server) as kafka_admin_client:
        return kafka_admin_client.list_topics()


def list_consumer_groups(kafka_server: KafkaServer) -> List[str]:
    """列出 Kafka 集群中所有的消费者组"""
    with ConnKafkaAdminClient(kafka_server) as kafka_admin_client:
        return [group_tuple[0] for group_tuple in kafka_admin_client.list_consumer_groups()]


def get_topic_configs(kafka_topic: KafkaTopic) -> Dict[str, Any]:
    """获取 TOPIC 的配置信息"""
    with ConnKafkaAdminClient(kafka_topic.kafka_server) as kafka_admin_client:
        configs = kafka_admin_client.describe_configs(config_resources=[
            ConfigResource(resource_type=ConfigResourceType.TOPIC, name=kafka_topic.topic)
        ], include_synonyms=False)

        topic_configs = {}

        config: DescribeConfigsResponse_v2
        for config in configs:
            for resource in config.get_item("resources"):
                _, _, _, _, config_entries = resource
                for config_name, config_value, _, _, _, _ in config_entries:
                    topic_configs[config_name] = config_value
        return topic_configs


def get_consume_topic(kafka_group: KafkaGroup) -> List[str]:
    """获取消费者组消费的 TOPIC 信息"""
    with ConnKafkaAdminClient(kafka_group.kafka_server) as kafka_admin_client:
        topic_set = set()
        for topic_partition in kafka_admin_client.list_consumer_group_offsets(kafka_group.group):
            topic_set.add(topic_partition.topic)
        return list(topic_set)



"""
连接器节点
"""

import dataclasses
from typing import List, Optional

from metasequoia_connector.node.common import HostPort
from metasequoia_connector.node.ssh_tunnel import SshTunnel


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class KafkaServer:
    """Kafka 集群对象"""

    bootstrap_servers: List[str] = dataclasses.field(kw_only=True)
    ssh_tunnel: Optional[SshTunnel] = dataclasses.field(kw_only=True, default=None)

    def get_host_list(self) -> List[HostPort]:
        return [HostPort.create_by_url(server) for server in self.bootstrap_servers]

    def __hash__(self):
        return hash((tuple(self.bootstrap_servers), self.ssh_tunnel))


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class KafkaTopic:
    """Kafka TOPIC"""

    kafka_server: KafkaServer = dataclasses.field(kw_only=True)
    topic: str = dataclasses.field(kw_only=True)


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class KafkaGroup:
    """Kafka 消费者组"""

    kafka_server: KafkaServer = dataclasses.field(kw_only=True)
    group: str = dataclasses.field(kw_only=True)


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class KafkaTopicsGroup:
    """Kafka TOPIC 并指定消费者组"""

    kafka_server: KafkaServer = dataclasses.field(kw_only=True)
    topics: List[str] = dataclasses.field(kw_only=True)
    group_id: str = dataclasses.field(kw_only=True)



import dataclasses
from typing import Tuple

__all__ = ["SshTunnel"]


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class SshTunnel:
    """SSH 隧道"""

    host: str = dataclasses.field(kw_only=True)  # 对应 ssh_address_or_host 参数
    port: int = dataclasses.field(kw_only=True)  # 对应 ssh_address_or_host 参数
    username: str = dataclasses.field(kw_only=True)
    password: str = dataclasses.field(kw_only=True, default=None)
    pkey: str = dataclasses.field(kw_only=True, default=None)

    @property
    def address(self) -> Tuple[str, int]:
        return self.host, self.port



"""
海豚元数据
"""

import dataclasses

from metasequoia_connector.node.mysql import MysqlInstance

__all__ = ["DSProcess", "DSTask", "DSProcessTask", "DSMetaInstance"]


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class DSProcess:
    """海豚工作流节点"""

    project_code: int = dataclasses.field(kw_only=True)
    process_code: int = dataclasses.field(kw_only=True)

    def get_process_url(self, domain: str):
        """获取工作流定义的 Url"""
        return f"{domain}/dolphinscheduler/ui/projects/{self.project_code}/workflow/definitions/{self.process_code}"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class DSTask:
    """海豚任务节点"""

    project_code: int = dataclasses.field(kw_only=True)
    task_code: int = dataclasses.field(kw_only=True)


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class DSProcessTask:
    """海豚工作流任务节点"""

    project_code: int = dataclasses.field(kw_only=True)
    process_code: int = dataclasses.field(kw_only=True)
    task_code: int = dataclasses.field(kw_only=True)

    def get_process_url(self, domain: str):
        """获取工作流定义的 Url"""
        return f"{domain}/dolphinscheduler/ui/projects/{self.project_code}/workflow/definitions/{self.process_code}"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class DSMetaInstance(MysqlInstance):
    """海豚调度元数据实例"""

    db: str = dataclasses.field(kw_only=True)



"""
OTS 节点
"""

import dataclasses

__all__ = ["OTSInstance", "OTSTable", "OTSIndex"]


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class OTSInstance:
    """OTS 实例"""

    end_point: str = dataclasses.field(kw_only=True)
    access_key_id: str = dataclasses.field(kw_only=True)
    access_key_secret: str = dataclasses.field(kw_only=True)
    instance_name: str = dataclasses.field(kw_only=True)


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class OTSTable:
    """OTS 表"""

    ots_instance: OTSInstance = dataclasses.field(kw_only=True)
    table_name: str = dataclasses.field(kw_only=True)

    def get_ots_index(self, index_name: str):
        """获取当前表中名为 index_name 的多元索引对象"""
        return OTSIndex(ots_instance=self.ots_instance, table_name=self.table_name, index_name=index_name)


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class OTSIndex:
    """OTS 多元索引"""

    ots_instance: OTSInstance = dataclasses.field(kw_only=True)
    table_name: str = dataclasses.field(kw_only=True)
    index_name: str = dataclasses.field(kw_only=True)



from typing import List, Optional, Tuple

import sshtunnel

from metasequoia_connector.node import SshTunnel

__all__ = ["create_ssh_tunnel"]


def create_ssh_tunnel(ssh_tunnel: SshTunnel,
                      remote_bind_address: Optional[Tuple[str, int]] = None,
                      remote_bind_addresses: Optional[List[Tuple[str, int]]] = None
                      ) -> sshtunnel.SSHTunnelForwarder:
    """使用 SshTunnel 隧道对象创建到 remote_host:remote_port 的 SSH 隧道

    TODO 待调整所有 SSH Tunnel 的调用逻辑，避免多个绑定地址报错

    关闭隧道时需要 close()
    建议使用 WITH ... AS ... 的语法调用这个函数

    Parameters
    ----------
    ssh_tunnel : SshTunnel
        SSH 隧道对象
    remote_bind_address : Optional[Tuple[str, int]], default = None
        远端服务器地址和端口的元组，用于连接到单个远端地址
    remote_bind_addresses: Optional[List[Tuple[str, int]]], default = None
        远端服务器地址和端口的元组的列表，用于连接到多个远端地址的列表
    """
    ssh_tunnel_forwarder = sshtunnel.SSHTunnelForwarder(
        ssh_address_or_host=(ssh_tunnel.host, ssh_tunnel.port),
        ssh_username=ssh_tunnel.username,
        ssh_password=ssh_tunnel.password,
        ssh_pkey=ssh_tunnel.pkey,
        remote_bind_address=remote_bind_address,
        remote_bind_addresses=remote_bind_addresses
    )
    ssh_tunnel_forwarder.start()
    return ssh_tunnel_forwarder



from metasequoia_connector.connector.common import WrapConnector
from metasequoia_connector.connector.ds_meta import DSMetaConnector
from metasequoia_connector.connector.hive import HiveConnector
from metasequoia_connector.connector.kafka import ConnKafkaAdminClient, ConnKafkaConsumer, ConnKafkaProducer
from metasequoia_connector.connector.mysql import MysqlConnector
from metasequoia_connector.connector.ots import OTSConnector
from metasequoia_connector.connector.redis import RedisConnectionPool, RedisConnector
from metasequoia_connector.connector.ssh_tunnel import create_ssh_tunnel



"""
OTS 连接器
"""

import ssl
from typing import Optional

import otssql
from metasequoia_connector.node import OTSInstance, SshTunnel

__all__ = ["OTSConnector"]


class OTSConnector:
    def __init__(self, ots_instance: OTSInstance, ssh_tunnel_info: Optional[SshTunnel] = None, **params) -> None:
        """OTS 连接的构造方法

        Parameters
        ----------
        ots_instance : OTSInstance
            OTS 实例的对象
        ssh_tunnel_info : Optional[SshTunnel]
            SSH 隧道的对象 TODO 暂未生效
        """
        self._ssh_tunnel_con = None
        self._is_close = False

        self._ots_con = otssql.connect(
            end_point=ots_instance.end_point,
            access_key_id=ots_instance.access_key_id,
            access_key_secret=ots_instance.access_key_secret,
            instance_name=ots_instance.instance_name,
            ssl_version=ssl.PROTOCOL_TLSv1_2,
            **params
        )

    def close(self, is_del: bool = False):
        if self._is_close is False:
            if self._ots_con is not None:
                # noinspection PyBroadException
                try:
                    self._ots_con.close()
                except Exception:
                    pass
            if self._ssh_tunnel_con is not None and is_del is False:
                # 如果是在对象销毁的过程中调用，则不销毁 SshTunnel 对象，因为该对象在销毁过程中调用 close 方法可能出现无法预料的报错或卡死
                # noinspection PyBroadException
                try:
                    self._ssh_tunnel_con.stop()
                except Exception:
                    pass

    def __getattr__(self, name):
        """代理嵌套连接类中的所有成员"""
        return getattr(self._ots_con, name)

    def __enter__(self):
        """在进入 WITH 语句时被调用（返回值作为 AS 后面的变量）"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """在离开 WITH 语句时被调用"""
        self.close()

    def __del__(self):
        """在对象被销毁时调用"""
        self.close(is_del=True)



"""
Hive 工具函数
"""

from pyhive import hive

from metasequoia_connector.connector import HiveConnector
from metasequoia_connector.node import HiveInstance


def execute(hive_instance: HiveInstance, sql: str):
    """执行 Hive 语句"""
    with HiveConnector(hive_instance) as conn:
        with conn.cursor() as cursor:
            result = cursor.execute(sql)
            return result


def execute_and_fetchall(hive_instance: HiveInstance, sql: str):
    """执行 Hive 语句"""
    with HiveConnector(hive_instance) as conn:
        with conn.cursor() as cursor:
            cursor.execute(sql)
            result = cursor.fetchall()
            return result


def conn_execute(hive_conn: hive.Connection, sql: str):
    """执行 Hive 语句"""
    with hive_conn.cursor() as cursor:
        result = cursor.execute(sql)
        return result


def conn_execute_and_fetchall(hive_conn: hive.Connection, sql: str):
    """执行 Hive 语句"""
    with hive_conn.cursor() as cursor:
        cursor.execute(sql)
        result = cursor.fetchall()
        return result


def conn_get_create_table(hive_conn: hive.Connection, table_name: str):
    """获取建表语句"""
    with hive_conn.cursor() as cursor:
        cursor.execute(f"SHOW CREATE TABLE {table_name}")
        result = []
        for row in cursor.fetchall():
            result.append(row[0])
        return "\n".join(result)



from distutils.core import setup

from setuptools import find_packages

with open("README.md", "r", encoding="UTF-8") as file:
    long_description = file.read()

setup(
    name="metasequoia-connector",
    version="0.0.5",
    description="水杉工具箱：数据库连接器",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="changxing",
    author_email="1278729001@qq.com",
    url="",
    install_requires=["otssql", "DBUtils", "PyHive", "PyMySQL", "setuptools", "sshtunnel", "metasequoia_sql"],
    license="MIT License",
    packages=find_packages(),
    platforms=["all"],
    classifiers=[
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Natural Language :: Chinese (Simplified)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Software Development :: Libraries"
    ]
)



"""
海豚调度的工具方法
"""

import collections
import datetime
import json
from typing import List, Dict, Any, Tuple, Set, Union, Optional, Iterable

import pymysql

from metasequoia_connector.node import DSProcess, DSProcessTask, DSTask
from metasequoia_connector.utils import mysql, sql_format

GeneralTask = Union[DSProcessTask, DSTask]


def get_project_list(conn: pymysql.Connection) -> Tuple[Dict[str, Any], ...]:
    """获取海豚调度的项目列表"""
    return mysql.conn_select_all_as_dict(
        conn, "SELECT `code`, `name` "
              "FROM t_ds_project"
    )


def get_process_list_by_project(conn: pymysql.Connection, project_code: str) -> Tuple[Dict[str, Any], ...]:
    """获取海豚调度指定项目的工作流列表"""
    return mysql.conn_select_all_as_dict(
        conn, f"SELECT `code`, `name` "
              f"FROM t_ds_process_definition "
              f"WHERE `project_code` = '{project_code}'"
    )


def get_task_list_by_process(conn: pymysql.Connection,
                             project_code: str,
                             process_code: str
                             ) -> Tuple[Dict[str, Any], ...]:
    """获取海豚调度指定项目、工作流的任务列表"""
    return mysql.conn_select_all_as_dict(
        conn, f"SELECT DISTINCT `post_task_code` "
              f"FROM t_ds_process_task_relation "
              f"WHERE `project_code` = '{project_code}' "
              f"  AND `process_definition_code`= '{process_code}'"
    )


def get_down_process_by_process(conn: pymysql.Connection,
                                process: DSProcess
                                ) -> List[DSProcess]:
    """根据 DSProcess 工作流对象，查询依赖该任务的下游工作流对象的列表"""
    query_result = mysql.conn_select_all_as_dict(
        conn, f"SELECT t1.project_code, t1.process_definition_code "
              f"FROM t_ds_process_task_relation AS t1 "
              f"INNER JOIN ("
              f"    SELECT `project_code`, `code` "
              f"    FROM t_ds_task_definition "
              f"    WHERE task_type = 'DEPENDENT' AND task_params LIKE '%{process.project_code}%{process.process_code}%'"
              f") AS t2 ON t1.project_code = t2.project_code AND t1.post_task_code = t2.code;"
    )
    return [DSProcess(project_code=query_row["project_code"], process_code=query_row["process_definition_code"])
            for query_row in query_result]


def get_dependent_task_params_by_task_list(conn: pymysql.Connection,
                                           project_code: str,
                                           task_code_list: List[int]) -> Tuple[Dict[str, Any], ...]:
    """获取海豚指定任务列表中依赖的其他工作流任务的任务参数"""
    task_code_str = ", ".join(f"'{task_code}'" for task_code in task_code_list)
    return mysql.conn_select_all_as_dict(
        conn, f"SELECT `task_params` "
              f"FROM t_ds_task_definition "
              f"WHERE `project_code` = '{project_code}' "
              f"  AND `code` IN ({task_code_str}) "
              f"  AND `task_type` ='DEPENDENT'"
    )


def get_task_code_list_by_process_code_list(conn: pymysql.Connection, project_code: int,
                                            process_code_list: List[int]) -> List[int]:
    """获取 project_code 中的 process_code_list 对应的 task_code 的列表"""
    process_code_list_str = ",".join([f"'{process_code}'" for process_code in process_code_list])
    return [row["post_task_code"] for row in mysql.conn_select_all_as_dict(
        conn, f"SELECT DISTINCT `post_task_code` "
              f"FROM t_ds_process_task_relation "
              f"WHERE `project_code` = '{project_code}' "
              f"  AND `process_definition_code` IN ({process_code_list_str})"
    )]


def _grouped_process_by_project(process_list: List[DSProcess]) -> Dict[int, List[DSProcess]]:
    """按所属项目对工作流进行分组"""
    grouped_process_dict = collections.defaultdict(list)
    for process in process_list:
        grouped_process_dict[process.project_code].append(process)
    return grouped_process_dict


def get_task_list_by_process_list(conn: pymysql.Connection,
                                  process_list: List[DSProcess]
                                  ) -> List[DSProcessTask]:
    """根据工作流集合，获取工作流包含的工作流任务节点列表"""
    result = []
    grouped_process_dict = _grouped_process_by_project(process_list)
    for project_code, grouped_process_list in grouped_process_dict.items():
        grouped_process_list_str = ",".join(str(process.process_code) for process in grouped_process_list)
        query_data = mysql.conn_select_all_as_dict(
            conn, f"SELECT DISTINCT `process_definition_code`, `post_task_code` "
                  f"FROM t_ds_process_task_relation "
                  f"WHERE `project_code` = '{project_code}' "
                  f"  AND `process_definition_code` IN ({grouped_process_list_str})"
        )
        for query_row in query_data:
            process_code = query_row["process_definition_code"]
            task_code = query_row["post_task_code"]
            result.append(DSProcessTask(project_code=project_code, process_code=process_code, task_code=task_code))
    return result


def get_dependence_process_set(conn: pymysql.Connection, process_list: List[DSProcess]) -> Set[DSProcess]:
    """获取当前任务依赖的所有上游任务的集合"""

    # 原始工作流集合
    source_process_set = set(process_list)

    # 已经添加过的工作流集合
    visited = set(process_list)
    while process_list:
        # 按所属项目对工作流进行分组
        grouped_process_dict = _grouped_process_by_project(process_list)

        # 清空旧的 process_list
        process_list.clear()

        # 按组分析工作流
        for project_code, group_process_list in grouped_process_dict.items():
            # 获取当前项目的工作流中包含的 task_code 的列表
            process_code_list = [process.process_code for process in group_process_list]
            task_code_list = get_task_code_list_by_process_code_list(conn, project_code, process_code_list)

            # 获取当前项目的 task_code_list 的列表中，DEPENDENT 类型 task 的 task_params 字段
            task_code_list_str = ", ".join(f"'{task_code}'" for task_code in task_code_list)
            dependent_tasks_list = mysql.conn_select_all_as_dict(
                conn, f"SELECT `task_params` "
                      f"FROM t_ds_task_definition "
                      f"WHERE `project_code` = '{project_code}' "
                      f"  AND `code` IN ({task_code_list_str}) "
                      f"  AND `task_type` ='DEPENDENT'"
            )

            # 遍历所有 task_params 字段，并将其中依赖的上游任务添加到新队列中
            for dependent_task in dependent_tasks_list:
                for dependence_task in json.loads(dependent_task["task_params"])["dependence"]["dependTaskList"]:
                    for dependence_item in dependence_task["dependItemList"]:
                        dependence_process = DSProcess(project_code=dependence_item["projectCode"],  # 依赖的项目ID
                                                       process_code=dependence_item["definitionCode"])  # 依赖的工作流ID
                        if dependence_process not in visited:
                            visited.add(dependence_process)
                            process_list.append(dependence_process)

    return visited - source_process_set


def get_schedule_release_state_by_process_list(conn: pymysql.Connection, process_list: List[DSProcess]
                                               ) -> Dict[DSProcess, int]:
    """获取当前工作流集合的定时上线状态"""
    process_hash = {process.process_code: process for process in process_list}
    process_list_str = ",".join(str(process.process_code) for process in process_list)
    query_data = mysql.conn_select_all_as_dict(conn, f"SELECT `process_definition_code`, `release_state` "
                                                     f"FROM `t_ds_schedules` "
                                                     f"WHERE `process_definition_code` IN ({process_list_str})")
    return {process_hash[query_row["process_definition_code"]]: query_row["release_state"] for query_row in query_data}


def get_schedule_priority_state_by_process_list(conn: pymysql.Connection, process_list: List[DSProcess]
                                                ) -> Dict[DSProcess, int]:
    """获取当前工作流集合的定时优先级"""
    process_hash = {process.process_code: process for process in process_list}
    process_list_str = ",".join(str(process.process_code) for process in process_list)
    query_data = mysql.conn_select_all_as_dict(conn, f"SELECT `process_definition_code`, `process_instance_priority` "
                                                     f"FROM `t_ds_schedules` "
                                                     f"WHERE `process_definition_code` IN ({process_list_str})")
    return {process_hash[query_row["process_definition_code"]]: query_row["process_instance_priority"]
            for query_row in query_data}


def get_process_name_dict_by_process_list(conn: pymysql.Connection,
                                          task_list: List[DSProcess]):
    """根据 process_list 获取 process 对象到任务名称的映射关系"""
    process_hash = {task.process_code: task for task in task_list}
    process_code_list_str = ",".join([f"{task.process_code}" for task in task_list])
    query_data = mysql.conn_select_all_as_dict(
        conn, f"SELECT `code`, `name` "
              f"FROM t_ds_process_definition "
              f"WHERE `code` IN ({process_code_list_str})"
    )
    return {process_hash[query_row["code"]]: query_row["name"] for query_row in query_data}


def get_task_name_dict_by_task_list(conn: pymysql.Connection,
                                    task_list: List[GeneralTask]):
    """根据 task_list 获取 task 对象到任务名称的映射关系"""
    task_hash = {task.task_code: task for task in task_list}
    task_code_list_str = ",".join([f"{task.task_code}" for task in task_list])
    query_data = mysql.conn_select_all_as_dict(
        conn, f"SELECT `code`, `name` "
              f"FROM t_ds_task_definition "
              f"WHERE `code` IN ({task_code_list_str})"
    )
    return {task_hash[query_row["code"]]: query_row["name"] for query_row in query_data}


def filter_process_task_by_condition(conn: pymysql.Connection,
                                     task_list: List[GeneralTask],
                                     condition: str) -> List[GeneralTask]:
    """根据 condition 条件过滤 task_list 中的任务，返回新的列表"""
    task_code_list_str = ",".join([f"{task.task_code}" for task in task_list])
    query_data = mysql.conn_select_all_as_dict(
        conn, f"SELECT `code` "
              f"FROM t_ds_task_definition "
              f"WHERE `code` IN ({task_code_list_str}) "
              f"  AND {condition}"
    )
    task_code_set = {query_row["code"] for query_row in query_data}
    return [process_task for process_task in task_list if process_task.task_code in task_code_set]


def get_process_info_by_process(conn: pymysql.Connection, process: DSProcess) -> dict:
    """获取海豚调度指定工作流的名称"""
    return mysql.conn_select_one_as_dict(
        conn,
        f"SELECT * "
        f"FROM t_ds_process_definition "
        f"WHERE `project_code` = '{process.project_code}' AND `code` = '{process.process_code}'"
    )


def get_process_info_by_process_list(conn: pymysql.Connection,
                                     process_list: List[DSProcess],
                                     need_columns: Optional[List[str]] = None) -> Iterable[Dict[str, Any]]:
    """使用 DSProcess 对象的列表，查询工作流定义的详细信息"""
    # 检查并规范化参数：必须返回 project_code 和 code
    if need_columns is not None:
        if "project_code" not in need_columns:
            need_columns.append("project_code")
        if "code" not in need_columns:
            need_columns.append("code")

    # 构造查询 SQL 并执行查询
    select_column_str = ",".join([f"`{column}`" for column in need_columns]) if need_columns is not None else "*"
    process_list_str = sql_format.to_quote_str_list_none_as_ignore([process.process_code for process in process_list])
    query_result = mysql.conn_select_all_as_dict(
        conn=conn,
        sql=f"SELECT {select_column_str} FROM t_ds_process_definition WHERE `code` IN {process_list_str}"
    )

    # 根据 project_code 筛选查询结果
    final_result = []
    process_hash = {process.process_code: process.project_code for process in process_list}
    for query_row in query_result:
        process_code = query_row["code"]
        project_code = query_row["project_code"]
        if project_code == process_hash[process_code]:
            final_result.append(query_row)
    return final_result


def get_scheduler_info_by_process_list(conn: pymysql.Connection,
                                       process_list: List[DSProcess],
                                       need_columns: Optional[List[str]] = None) -> Iterable[Dict[str, Any]]:
    """使用 DSProcess 对象的列表，查询工作流的定时的上线状态"""
    select_column_str = ",".join([f"`{column}`" for column in need_columns]) if need_columns is not None else "*"
    process_list_str = sql_format.to_quote_str_list_none_as_ignore([process.process_code for process in process_list])
    return mysql.conn_select_all_as_dict(
        conn=conn,
        sql=f"SELECT {select_column_str} FROM t_ds_schedules WHERE process_definition_code IN {process_list_str}"
    )


def get_recent_state_history_by_process(conn: pymysql.Connection,
                                        process: DSProcess,
                                        limit: int = 30) -> List[str]:
    """获取 process 工作流最近 limit 次运行的 state_history 字段"""
    query_result = mysql.conn_select_all_as_dict(
        conn=conn,
        sql=f"SELECT state_history "
            f"FROM t_ds_process_instance "
            f"WHERE process_definition_code = '{process.process_code}' "
            f"ORDER BY id DESC LIMIT {limit}"
    )
    return [query_row["state_history"] for query_row in query_result]


def get_recent_start_time_and_success_time_by_process(
        conn: pymysql.Connection,
        process: DSProcess,
        limit: int = 30
) -> List[Dict[str, Optional[datetime.datetime]]]:
    """获取 process 工作流最近 limit 次运行的开始时间和结束时间"""
    result = []
    for state_history_source in get_recent_state_history_by_process(conn, process, limit):
        state_history = json.loads(state_history_source)
        row = {}
        for state_item in state_history:
            state_time = datetime.datetime.strptime(state_item["time"], "%Y-%m-%d %H:%M:%S")
            if (state_item["state"] == "RUNNING_EXECUTION"
                    and state_item["desc"] == "init running"):
                row["init_time"] = state_time  # 初始化时间
            if (state_item["state"] == "RUNNING_EXECUTION"
                    and state_item["desc"] == "start a new process from scheduler"):
                row["start_process_time"] = state_time  # 进程启动时间
            if state_item["state"] == "SUCCESS":
                row["success_time"] = state_time  # 成功时间
        result.append(row)
    return result



"""
组合节点
"""

from metasequoia_connector.node.kafka import KafkaTopic
from metasequoia_connector.node.mysql import MysqlTable

__all__ = ["MysqlTableWithKafkaTopic"]


class MysqlTableWithKafkaTopic:
    """有 Kafka TOPIC 监听的 RDS 表"""

    def __init__(self, rds_table: MysqlTable, kafka_topic: KafkaTopic):
        self._rds_table = rds_table
        self._kafka_topic = kafka_topic

    @property
    def rds_table(self) -> MysqlTable:
        return self._rds_table

    @property
    def kafka_topic(self) -> KafkaTopic:
        return self._kafka_topic



from metasequoia_connector.node.combine import MysqlTableWithKafkaTopic
from metasequoia_connector.node.common import HostPort
from metasequoia_connector.node.common import HostPort
from metasequoia_connector.node.ds_meta import DSMetaInstance, DSProcess, DSProcessTask, DSTask
from metasequoia_connector.node.hive import HiveInstance, HiveTable
from metasequoia_connector.node.kafka import KafkaGroup, KafkaServer, KafkaTopic, KafkaTopicsGroup
from metasequoia_connector.node.mysql import MysqlInstance, MysqlTable
from metasequoia_connector.node.ots import OTSIndex, OTSInstance, OTSTable
from metasequoia_connector.node.redis import RedisDatabase, RedisInstance
from metasequoia_connector.node.ssh_tunnel import SshTunnel
